import numpy as np

from .points import get_fractions, interpolate_along_axis


def prepare(start: np.ndarray, stop: np.ndarray):
    start = normalize(start)
    stop = normalize(stop)

    diff = len(start) - len(stop)
    if diff < 0:
        start = add_points(start, -diff)
    elif diff > 0:
        stop = add_points(stop, diff)

    start = rotate(start, stop)
    return start, stop


def get_area(line):
    line = np.insert(line, 0, line[-1], axis=0)
    return np.cross(line[1:], line[:-1]).sum() / 2


def normalize(line):
    if line.shape[1] != 2:
        raise ValueError('In case of closed lines only 2D points are supported.')

    if get_area(line) > 0:
        line = line[::-1]

    return line


def rotate(line, reference):
    min_, best_offset = np.inf, 0

    for offset in range(len(line)):
        deviation = np.linalg.norm(reference - np.roll(line, offset, axis=0))

        if deviation < min_:
            min_ = deviation
            best_offset = offset

    if best_offset:
        line = np.roll(line, best_offset, axis=0)

    return line


def add_points(line, n_points):
    line = np.insert(line, len(line), line[0], axis=0)
    fractions = get_fractions(line)
    new_fractions = sorted(
        list(fractions) +
        list(np.linspace(0, 1, n_points, endpoint=False) + 1 / (2 * n_points))
    )[:-1]

    return interpolate_along_axis(new_fractions, fractions, line)



import numpy as np


def get_fractions(line):
    fractions = np.linalg.norm(np.diff(line, axis=0), axis=1).cumsum()
    # if the line is made of the same point
    if fractions[-1] == 0:
        fractions[-1] = 1

    fractions /= fractions[-1]
    return np.insert(fractions, 0, 0)


def interpolate_along_axis(new_coords, coords, values):
    return np.stack([np.interp(new_coords, coords, v) for v in values.T], 1)



from typing import Union, Iterable, Sequence, Callable

import numpy as np

from .closed import prepare as prepare_closed
from .open import prepare as prepare_open

Line = Union[np.ndarray, Iterable[Sequence[float]]]


def _normalize_parameter(t):
    if not (0 <= t <= 1):
        raise ValueError('The interpolation parameter must be between 0 and 1.')

    return t


def _normalize_line(line):
    assert len(line) > 0
    line = np.asarray(line)
    assert line.ndim == 2
    # TODO: remove duplicates
    return line


def interpolate(start: Line, stop: Line, t: float, closed: bool = True) -> np.ndarray:
    """
    Interpolate between ``start`` and ``stop`` according to interpolation parameter ``t``.

    Parameters
    ----------
    start: numpy.ndarray
        the starting line. Expected shape (N, d),
        where N is the number of points and d - their dimensionality.
    stop: numpy.ndarray
        the ending line. Expected shape (M, d),
        where M is the number of points and d - their dimensionality.
    t: float
        the interpolation parameter. Must be between 0 and 1.
    closed: bool
        whether the lines are closed

    Returns
    -------
    interpolated_line: numpy.ndarray

    Notes
    -----
    The number of points in each line are not necessarily equal.
    In case of closed lines only 2D points are supported.
    """
    t = _normalize_parameter(t)
    return interpolator(start, stop, closed)(t)


def interpolator(start: Line, stop: Line, closed: bool = True) -> Callable:
    """
    Returns a function that interpolates between ``start`` and ``stop`` given an interpolation parameter.

    Parameters
    ----------
    start: numpy.ndarray
        the starting line. Expected shape (N, d),
        where N is the number of points and d - their dimensionality.
    stop: numpy.ndarray
        the ending line. Expected shape (M, d),
        where M is the number of points and d - their dimensionality.
    closed: bool
        whether the lines are closed

    Returns
    -------
    interpolator: a function that given the interpolation parameter t
        returns the interpolated line.
    """

    def interpolate_(t: float):
        t = _normalize_parameter(t)
        return start + delta * t

    prepare = prepare_closed if closed else prepare_open
    start, stop = prepare(_normalize_line(start), _normalize_line(stop))
    delta = stop - start

    return interpolate_



import numpy as np
from .points import get_fractions, interpolate_along_axis


def prepare(start: np.ndarray, stop: np.ndarray):
    # handling a possible flip
    diff = np.linalg.norm(start[[0, -1]] - stop[[0, -1]], axis=1).sum()
    rev_diff = np.linalg.norm(start[[-1, 0]] - stop[[0, -1]], axis=1).sum()
    if rev_diff < diff:
        start = start[::-1]

    start_fractions = get_fractions(start)
    stop_fractions = get_fractions(stop)
    merged = np.union1d(start_fractions, stop_fractions)

    start = interpolate_along_axis(merged, start_fractions, start)
    stop = interpolate_along_axis(merged, stop_fractions, stop)
    return start, stop



import runpy
from pathlib import Path

from setuptools import setup, find_packages

classifiers = '''Development Status :: 5 - Production/Stable
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9'''

name = 'pyflubber'
root = Path(__file__).parent
with open(root / 'README.md', encoding='utf-8') as file:
    long_description = file.read()
with open(root / 'requirements.txt', encoding='utf-8') as file:
    requirements = file.read().splitlines()
version = runpy.run_path(root / name / '__version__.py')['__version__']

setup(
    name=name,
    packages=find_packages(include=(name,)),
    include_package_data=True,
    version=version,
    descriprion='Linear interpolation of open and closed lines.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/maxme1/pyflubber',
    download_url='https://github.com/maxme1/pyflubber/v%s.tar.gz' % version,
    keywords=['interpolation', 'curves'],
    classifiers=classifiers.splitlines(),
    install_requires=requirements,
)
